
type 'a set =
	| Empty
	| Fork of 'a node 
and 'a node =
	{left:'a set;item:'a;right:'a set}

(* strict catamorphism *)
type ('a,'b) fold =
	<empty: 'a; fork: 'a -> 'b -> 'a -> 'a>
let rec fold s (case:('a,'b) fold) =
	match s with
	| Empty -> case#empty
	| Fork n -> case#fork (fold n.left case) n.item (fold n.right case)
(* lazy catamorphism *)
type ('a,'b) cata =
	<empty: 'a; fork: (unit -> 'a) -> 'b -> (unit -> 'a) -> 'a>
let rec cata s (case:('a,'b) cata) () =
	match s with
	| Empty -> case#empty
	| Fork n -> case#fork (cata n.left case) n.item (cata n.right case)

(* strict paramorphism *)
type ('a,'b) recu =
	<empty: 'b;
	 fork: 'a node -> 'b -> 'b -> 'b>
let rec recu s (case:('a,'b) recu) =
	match s with
	| Empty ->
			case#empty
	| Fork n ->
			case#fork n (recu n.left case) (recu n.right case)
(* lazy paramorphism *)
type ('a,'b) para =
	<empty: 'b;
	 fork: 'a node -> (unit -> 'b) -> (unit -> 'b) -> 'b>
let rec para s (case:('a,'b) para) () =
	match s with
	| Empty ->
			case#empty
	| Fork n ->
			case#fork n (para n.left case) (para n.right case)


let strahler s =
	fold s
		object
			method empty = 0
			method fork l _ r =
				if l = r then r+1 else max l r
		end
let cardinal s =
	fold s
		object
			method empty = 0
			method fork l _ r = l + 1 + r
		end
let singleton item =
	Fork {left=Empty;item;right=Empty}
let add x s =
	para s
		object
			method empty =
				singleton x
			method fork n l r =
				let y = n.item in
				if x < y then Fork {n with left = l ()}
				else if x > y then Fork {n with right = r ()}
				else Fork n
		end ()
let member x s =
	cata s
		object
			method empty = false
			method fork l y r =
				if x < y then l()
				else if x > y then r()
				else true
		end ()

let interval low high s =
	cata s
		object
			method empty = Empty
			method fork l x r =
				if high < x then l ()
				else if low > x then r()
				else Fork {left=l ();item=x;right=r ()}
		end ()

let minimum n =
	fold n.left
		object
			method empty acc = acc
			method fork l x r acc = l x 
		end n.item

let remove_minimum n =
	recu n.left
		object
			method empty acc = acc
			method fork n l r acc = Fork {n with left=l n.right}
		end n.right
	
(* concatenation of sa + sb where max(sa) < min(sb) *)				
let concat sa sb =
	if sa = Empty then sb else
	match sb with
	| Empty -> sa
	| Fork n ->
		let item = minimum n
		and right = remove_minimum n
		in Fork {left=sa;item;right}

let remove x s =
	para s
		object
			method empty =
				Empty
			method fork n l r =
				if x < n.item then Fork {n with left=l()}
				else if x > n.item then Fork {n with right=r()}
				else concat n.left n.right
		end ()	
	
let split x s =
	para s
		object
			method empty =
				Empty,false,Empty
			method fork n l r =
				let y = n.item in
				if x < y then
					let a,b,c = l() in
					a,b,Fork {n with left = c}
				else if x > y then
					let a,b,c = r() in
					Fork {n with right = a},b,c
				else
					n.left,true,n.right
		end ()
let for_all cond s =
	cata s
		object
			method empty = true
			method fork l y r =
				l() && cond y && r()
		end ()

(* set operations *)
let union sa sb =
	fold sa
		object
			method empty s =
				s
			method fork la na ra s =
				match s with
				| Empty ->
					sa
				| Fork _ ->
					let lb,_,rb = split na s in
					Fork {left=la lb;item=na;right=ra rb}
		end sb
let intersection sa sb =
	fold sa
		object
			method empty s =
				Empty
			method fork la na ra s =
				match s with
				| Empty ->
					Empty
				| Fork _ ->
					let lb,b,rb = split na s in
					if b then Fork {left=la lb;item=na;right=ra rb}
					else concat (la lb) (ra rb)
		end sb
let difference sa sb =
	recu sa
		object
			method empty s =
				Empty
			method fork n la ra s =
				match s with
				| Empty ->
					Fork n
				| Fork _ ->
					let lb,b,rb = split n.item s in
					if b then concat (la lb) (ra rb)
					else Fork {left=la lb;item=n.item;right=ra rb}
		end sb
let disjoint sa sb =
	fold sa
		object
			method empty s =
				true
			method fork la na ra s =
				match s with
				| Empty ->
					true
				| Fork _ ->
					let lb,b,rb = split na s in
					if b then false
					else la lb && ra rb
		end sb
let subset sa sb =
	fold sa
		object
			method empty s =
				s = Empty
			method fork la na ra s =
				match s with
				| Empty ->
					false
				| Fork _ ->
					let lb,b,rb = split na s in
					if b then la lb && ra rb
					else false
		end sb

let equal sa sb =
	subset sa sb && subset sb sa
